/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <stddef.h>
#include <stdint.h>
#include <spu_mfcio.h>

#include "config.h"

#include "mars/error.h"
#include "mars/module.h"

#include "kernel_internal_types.h"
#include "workload_internal_types.h"

/* kernel */
void *__kernel_stack;
static struct mars_kernel_params kernel_params;

/* workload queue */
static struct mars_workload_queue_header queue_header;
static struct mars_workload_queue_block queue_block;

/* workload */
static struct mars_workload_context workload;
static uint16_t workload_id;
static uint64_t workload_ea;
static uint8_t workload_state;
static uint8_t workload_cached;

/* workload to schedule */
static struct mars_workload_context schedule_workload;
static uint16_t schedule_workload_id;

/* workload module cached */
static struct mars_workload_module cached_workload_module = {0, 0, 0, 0};

/* workload module entry */
typedef void (*module_entry)(
	const struct mars_kernel_syscalls *kernel_syscalls);

static int kernel_memcmp(const void *s1, const void *s2, int size)
{
	unsigned char *ptr_1 = (unsigned char *)s1;
	unsigned char *ptr_2 = (unsigned char *)s2;
	unsigned char *ptr_e = (unsigned char *)ptr_1 + size;

	while (ptr_1 < ptr_e) {
		if (*ptr_1 != *ptr_2)
			return 1;
		ptr_1++;
		ptr_2++;
	}

	return 0;
}

static void kernel_memcpy(void *dst, const void *src, int size)
{
	unsigned char *ptr_1 = (unsigned char *)dst;
	unsigned char *ptr_2 = (unsigned char *)src;
	unsigned char *ptr_e = (unsigned char *)ptr_1 + size;

	while (ptr_1 < ptr_e) {
		*ptr_1 = *ptr_2;
		ptr_1++;
		ptr_2++;
	}
}

static uint32_t get_ticks(void)
{
	return kernel_params.kernel_ticks.offset - spu_read_decrementer();
}

static uint64_t get_mars_context_ea(void)
{
	return kernel_params.mars_context_ea;
}

static uint16_t get_kernel_id(void)
{
	return kernel_params.kernel_id;
}

static uint16_t get_workload_id(void)
{
	return workload_id;
}

static struct mars_workload_context *get_workload(void)
{
	return &workload;
}

static uint64_t get_workload_ea(uint16_t id)
{
	return queue_header.context_ea +
	       id * sizeof(struct mars_workload_context);
}

static struct mars_workload_context *get_workload_by_id(uint16_t id)
{
	static struct mars_workload_context ret_workload;

	/* id is caller workload's id so return current workload */
	if (id == workload_id)
		return &workload;

	/* get the workload context from workload queue */
	dma_get((void *)&ret_workload, get_workload_ea(id),
		sizeof(struct mars_workload_context), MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);

	return &ret_workload;
}

static void __attribute__((noinline)) save_workload_state(uint8_t state)
{
	workload_state = state;
}

static void workload_exit(uint8_t state)
{
	register void *sp asm("$sp");

	save_workload_state(state);

	/* restore kernel stack pointer */
	sp = __kernel_stack;
}

static uint64_t get_block_ea(int block)
{
	return queue_header.queue_ea +
	       offsetof(struct mars_workload_queue, block) +
	       sizeof(struct mars_workload_queue_block) * block;
}

static uint64_t get_block_bits(uint16_t id)
{
	int block;
	int index;
	uint64_t block_ea;
	uint64_t block_bits;

	/* check function params */
	if (id >= MARS_WORKLOAD_MAX)
		return 0;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = get_block_ea(block);

	/* lock the queue block */
	mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	block_bits = queue_block.bits[index];

	/* unlock the queue block */
	mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return block_bits;
}

static void update_header_bits(int block)
{
	int i;
	uint16_t *block_bits = &queue_header.bits[block];
	uint8_t block_ready = MARS_WORKLOAD_BLOCK_READY_OFF;
	uint8_t block_waiting = MARS_WORKLOAD_BLOCK_WAITING_OFF;
	uint8_t block_priority = MARS_WORKLOAD_BLOCK_PRIORITY_MIN;

	/* search through currently locked queue block workload bits */
	for (i = 0; i < MARS_WORKLOAD_PER_BLOCK; i++) {
		uint64_t *bits = &queue_block.bits[i];
		uint8_t state = MARS_BITS_GET(bits, WORKLOAD_STATE);

		/* workload state is ready so check priority */
		if (state == MARS_WORKLOAD_STATE_READY) {
			uint8_t priority = MARS_BITS_GET(bits,
							 WORKLOAD_PRIORITY);

			/* set block priority if higher then current */
			if (priority > block_priority)
				block_priority = priority;

			/* set block ready bit in header bits for block */
			block_ready = MARS_WORKLOAD_BLOCK_READY_ON;
		} else if (state == MARS_WORKLOAD_STATE_WAITING) {
			/* set block waiting bit in header bits for block */
			block_waiting = MARS_WORKLOAD_BLOCK_WAITING_ON;
		}
	}

	/* lock the queue header */
	mutex_lock_get(kernel_params.workload_queue_ea,
		       (struct mars_mutex *)&queue_header);

	/* set the info bits inside queue header for this queue block */
	MARS_BITS_SET(block_bits, BLOCK_READY, block_ready);
	MARS_BITS_SET(block_bits, BLOCK_WAITING, block_waiting);
	MARS_BITS_SET(block_bits, BLOCK_PRIORITY, block_priority);

	/* increment the header access value */
	queue_header.access++;

	/* unlock the queue header */
	mutex_unlock_put(kernel_params.workload_queue_ea,
			 (struct mars_mutex *)&queue_header);
}

static void update_header_bits_counter(int block, int reset)
{
	uint16_t *block_bits = &queue_header.bits[block];
	uint8_t block_counter = MARS_WORKLOAD_BLOCK_COUNTER_MIN;

	/* lock the queue header */
	mutex_lock_get(kernel_params.workload_queue_ea,
		       (struct mars_mutex *)&queue_header);

	/* reset is not specified so increment current block counter */
	if (!reset) {
		block_counter = MARS_BITS_GET(block_bits, BLOCK_COUNTER);
		if (block_counter < MARS_WORKLOAD_BLOCK_COUNTER_MAX)
			block_counter++;
	}

	/* set the block counter bits */
	MARS_BITS_SET(block_bits, BLOCK_COUNTER, block_counter);

	/* unlock the queue header */
	mutex_unlock_put(kernel_params.workload_queue_ea,
			 (struct mars_mutex *)&queue_header);
}

static int workload_query(uint16_t id, int query)
{
	uint64_t bits = get_block_bits(id);

	switch (query) {
	case MARS_QUERY_IS_CACHED:
		return (id == workload_id && workload_cached);
	case MARS_QUERY_IS_INITIALIZED:
		return (MARS_BITS_GET(&bits, WORKLOAD_STATE) !=
			MARS_WORKLOAD_STATE_NONE);
	case MARS_QUERY_IS_READY:
		return (MARS_BITS_GET(&bits, WORKLOAD_STATE) ==
			MARS_WORKLOAD_STATE_READY);
	case MARS_QUERY_IS_WAITING:
		return (MARS_BITS_GET(&bits, WORKLOAD_STATE) ==
			MARS_WORKLOAD_STATE_WAITING);
	case MARS_QUERY_IS_RUNNING:
		return (MARS_BITS_GET(&bits, WORKLOAD_STATE) ==
			MARS_WORKLOAD_STATE_RUNNING);
	case MARS_QUERY_IS_FINISHED:
		return (MARS_BITS_GET(&bits, WORKLOAD_STATE) ==
			MARS_WORKLOAD_STATE_FINISHED);
	case MARS_QUERY_IS_SIGNAL_SET:
		return (MARS_BITS_GET(&bits, WORKLOAD_SIGNAL) ==
			MARS_WORKLOAD_SIGNAL_ON);
	}

	return MARS_ERROR_PARAMS;
}

static int change_bits(uint16_t id,
		       int (*check_bits)(uint64_t bits, uint64_t param),
		       uint64_t check_bits_param,
		       uint64_t (*set_bits)(uint64_t bits, uint64_t param),
		       uint64_t set_bits_param,
		       void (*callback)(uint16_t id))
{
	int block;
	int index;
	uint64_t block_ea;
	uint64_t bits;

	/* check function params */
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = get_block_ea(block);

	/* lock the queue block */
	mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* check for valid state */
	if (check_bits &&
	    !(*check_bits)(queue_block.bits[index], check_bits_param)) {
		mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	/* reset workload queue bits and set state to new state */
	bits = (*set_bits)(queue_block.bits[index], set_bits_param);

	/* store new bits into queue block */
	queue_block.bits[index] = bits;

	/* if callback requested call it */
	if (callback)
		(*callback)(id);

	/* unlock the queue block */
	mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return MARS_SUCCESS;
}

static int check_state_bits(uint64_t bits, uint64_t state)
{
	return (MARS_BITS_GET(&bits, WORKLOAD_STATE) == state);
}

static int check_state_bits_not(uint64_t bits, uint64_t state)
{
	return (MARS_BITS_GET(&bits, WORKLOAD_STATE) != state);
}

static uint64_t set_state_bits(uint64_t bits, uint64_t state)
{
	MARS_BITS_SET(&bits, WORKLOAD_STATE, state);

	return bits;
}

static int change_state(uint16_t id,
			unsigned int old_state,
			unsigned int new_state,
			void (*callback)(uint16_t id))
{
	return change_bits(id,
			   check_state_bits, old_state,
			   set_state_bits, new_state,
		           callback);
}

static uint64_t set_wait_id_bits(uint64_t bits, uint64_t id)
{
	MARS_BITS_SET(&bits, WORKLOAD_WAIT_ID, id);

	return bits;
}

static int workload_wait_set(uint16_t id)
{
	if (id == workload_id)
		return MARS_ERROR_PARAMS;

	return change_bits(workload_id,
			   check_state_bits_not, MARS_WORKLOAD_STATE_NONE,
			   set_wait_id_bits, id,
			   NULL);
}

static int workload_wait_reset(void)
{
	return change_bits(workload_id,
			   NULL, 0,
			   set_wait_id_bits, MARS_WORKLOAD_ID_NONE,
			   NULL);
}

static uint64_t set_signal_bits(uint64_t bits, uint64_t signal)
{
	MARS_BITS_SET(&bits, WORKLOAD_SIGNAL, signal);

	return bits;
}

static int workload_signal_set(uint16_t id)
{
	if (id == workload_id)
		return MARS_ERROR_PARAMS;

	return change_bits(id,
			   check_state_bits_not, MARS_WORKLOAD_STATE_NONE,
			   set_signal_bits, MARS_WORKLOAD_SIGNAL_ON,
			   NULL);
}

static int workload_signal_reset(void)
{
	return change_bits(workload_id,
			   NULL, 0,
			   set_signal_bits, MARS_WORKLOAD_SIGNAL_OFF,
			   NULL);
}

static uint64_t set_schedule_bits(uint64_t bits, uint64_t priority)
{
	/* set the info bits inside queue block for this workload */
	MARS_BITS_SET(&bits, WORKLOAD_STATE, MARS_WORKLOAD_STATE_SCHEDULING);
	MARS_BITS_SET(&bits, WORKLOAD_PRIORITY, priority);
	MARS_BITS_SET(&bits, WORKLOAD_COUNTER, MARS_WORKLOAD_COUNTER_MIN);
	MARS_BITS_SET(&bits, WORKLOAD_SIGNAL, MARS_WORKLOAD_SIGNAL_OFF);
	MARS_BITS_SET(&bits, WORKLOAD_WAIT_ID, MARS_WORKLOAD_ID_NONE);

	return bits;
}

static void schedule_begin_callback(uint16_t id)
{
	/* get id of workload to schedule */
	schedule_workload_id = id;

	/* get the workload context from workload queue */
	dma_get((void *)&schedule_workload, get_workload_ea(id),
		sizeof(struct mars_workload_context), MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);
}

static int workload_schedule_begin(uint16_t id, uint8_t priority,
				   struct mars_workload_context **workload)
{
	int ret;

	/* check function params */
	if (id == workload_id)
		return MARS_ERROR_PARAMS;

	/* change bits necessary to begin scheduling */
	ret = change_bits(id,
			  check_state_bits, MARS_WORKLOAD_STATE_FINISHED,
			  set_schedule_bits, priority,
			  schedule_begin_callback);
	if (ret != MARS_SUCCESS)
		return ret;

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &schedule_workload;

	return MARS_SUCCESS;
}

static void schedule_end_callback(uint16_t id)
{
	/* get id of workload to schedule */
	schedule_workload_id = MARS_WORKLOAD_ID_NONE;

	/* put the workload context into workload queue */
	dma_put((void *)&schedule_workload, get_workload_ea(id),
		sizeof(struct mars_workload_context), MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);

	/* update queue header bits */
	update_header_bits(id / MARS_WORKLOAD_PER_BLOCK);
}

static int workload_schedule_end(uint16_t id)
{
	if (id != schedule_workload_id)
		return MARS_ERROR_PARAMS;

	return change_state(id,
			    MARS_WORKLOAD_STATE_SCHEDULING,
			    MARS_WORKLOAD_STATE_READY,
			    schedule_end_callback);
}

static int workload_schedule_cancel(uint16_t id)
{
	if (id != schedule_workload_id)
		return MARS_ERROR_PARAMS;

	return change_state(id,
			    MARS_WORKLOAD_STATE_SCHEDULING,
			    MARS_WORKLOAD_STATE_FINISHED,
			    NULL);
}

static int host_signal_send(uint64_t watch_point_ea)
{
#ifdef ENABLE_COND_WAIT_FUTEX
	spu_write_out_mbox((uint32_t)(watch_point_ea >> 32));
	spu_write_out_intr_mbox((uint32_t)(watch_point_ea & 0xffffffff));
#endif

	return MARS_SUCCESS;
}

static struct mars_kernel_syscalls kernel_syscalls =
{
	get_ticks,
	get_mars_context_ea,
	get_kernel_id,
	get_workload_id,
	get_workload,
	get_workload_by_id,
	workload_exit,
	workload_query,
	workload_wait_set,
	workload_wait_reset,
	workload_signal_set,
	workload_signal_reset,
	workload_schedule_begin,
	workload_schedule_end,
	workload_schedule_cancel,
	host_signal_send,
	mutex_lock_get,
	mutex_unlock_put,
	dma_get,
	dma_put,
	dma_wait
};

static int search_block(int block, int ready)
{
	int i;
	int index = -1;
	uint8_t max_priority = 0;
	uint16_t max_counter = 0;
	uint64_t block_ea = get_block_ea(block);

	/* lock the queue block */
	mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* search through all workloads in block */
	for (i = 0; i < MARS_WORKLOAD_PER_BLOCK; i++) {
		uint64_t *bits   = &queue_block.bits[i];
		uint8_t state    = MARS_BITS_GET(bits, WORKLOAD_STATE);
		uint8_t priority = MARS_BITS_GET(bits, WORKLOAD_PRIORITY);
		uint8_t signal   = MARS_BITS_GET(bits, WORKLOAD_SIGNAL);
		uint16_t wait_id = MARS_BITS_GET(bits, WORKLOAD_WAIT_ID);
		uint16_t counter = MARS_BITS_GET(bits, WORKLOAD_COUNTER);

		/* found workload in ready state */
		if (ready && state == MARS_WORKLOAD_STATE_READY) {
			/* compare priority and counter with previous ones */
			if (index < 0 || priority > max_priority ||
			  (priority == max_priority && counter > max_counter)) {
				index = i;
				max_counter = counter;
				max_priority = priority;
			}

			/* increment wait counter without overflowing */
			if (counter < MARS_WORKLOAD_COUNTER_MAX)
				MARS_BITS_SET(bits, WORKLOAD_COUNTER,
					      counter + 1);
		/* found workload in waiting state */
		} else if (!ready && state == MARS_WORKLOAD_STATE_WAITING) {
			/* waiting for workload to finish so check status */
			if (wait_id != MARS_WORKLOAD_ID_NONE) {
				struct mars_workload_queue_block wait_block;
				struct mars_workload_queue_block *p_wait_block;
				uint8_t wait_state;

				int bl = wait_id / MARS_WORKLOAD_PER_BLOCK;
				int id = wait_id % MARS_WORKLOAD_PER_BLOCK;

				/* check if workload id is in the same block */
				if (block != bl) {
					/* fetch the necessary block */
					dma_get(&wait_block, get_block_ea(bl),
						sizeof(wait_block),
						MARS_KERNEL_DMA_TAG);
					dma_wait(MARS_KERNEL_DMA_TAG);

					/* set pointer to check fetched block */
					p_wait_block = &wait_block;
				} else {
					/* set pointer to check current block */
					p_wait_block = &queue_block;
				}

				/* get state of workload its waiting for */
				wait_state =
					MARS_BITS_GET(&p_wait_block->bits[id],
						      WORKLOAD_STATE);

				/* check if workload is finished and reset */
				if (wait_state ==
				    MARS_WORKLOAD_STATE_FINISHED) {
					MARS_BITS_SET(bits, WORKLOAD_WAIT_ID,
						MARS_WORKLOAD_ID_NONE);
					MARS_BITS_SET(bits, WORKLOAD_STATE,
						MARS_WORKLOAD_STATE_READY);

					/* update queue header bits */
					update_header_bits(block);
				}
			/* waiting for signal so check signal bit and reset */
			} else if (signal == MARS_WORKLOAD_SIGNAL_ON) {
				MARS_BITS_SET(bits, WORKLOAD_SIGNAL,
					      MARS_WORKLOAD_SIGNAL_OFF);
				MARS_BITS_SET(bits, WORKLOAD_STATE,
					      MARS_WORKLOAD_STATE_READY);

				/* update queue header bits */
				update_header_bits(block);
			}
		}
	}

	/* index is set so reserve the runnable workload */
	if (index >= 0) {
		/* update the current state of the workload */
		MARS_BITS_SET(&queue_block.bits[index], WORKLOAD_STATE,
			      MARS_WORKLOAD_STATE_RUNNING);

		/* reset the counter for reserved workload */
		MARS_BITS_SET(&queue_block.bits[index], WORKLOAD_COUNTER,
			      MARS_WORKLOAD_COUNTER_MIN);

		/* check if this kernel ran this workload most recently */
		workload_cached =
			(workload_id ==
				MARS_WORKLOAD_PER_BLOCK * block + index) &&
			(kernel_params.kernel_id ==
				MARS_BITS_GET(&queue_block.bits[index],
					      WORKLOAD_KERNEL_ID));

		/* set the kernel id of this kernel into block bits */
		MARS_BITS_SET(&queue_block.bits[index], WORKLOAD_KERNEL_ID,
			      kernel_params.kernel_id);

		/* update queue header bits and reset block counter */
		update_header_bits(block);
		update_header_bits_counter(block, 1);
	}

	/* unlock the queue block */
	mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	/* returns -1 if no runnable workload found */
	return index;
}

static void notify_host_bits(uint64_t block_ea, int index)
{
	uint64_t bits_ea =
		block_ea +
		offsetof(struct mars_workload_queue_block, bits) +
		sizeof(uint64_t) * index;

	host_signal_send(bits_ea);
}

static int reserve_workload(void)
{
	int i;
	int block = -1;
	int index = -1;
	uint8_t max_block_priority = 0;
	uint16_t max_block_counter = 0;

	/* get the workload queue header */
	dma_get(&queue_header, kernel_params.workload_queue_ea,
		sizeof(struct mars_workload_queue_header), MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);

	/* return exit status if exit flag is set from host */
	if (queue_header.flag == MARS_WORKLOAD_QUEUE_FLAG_EXIT)
		return MARS_KERNEL_STATUS_EXIT;

	/* search workload queue header for highest priority ready block that
	 * has waited the longest in ready state */
	for (i = 0; i < MARS_WORKLOAD_NUM_BLOCKS; i++) {
		uint16_t *bits = &queue_header.bits[i];
		uint8_t block_ready    = MARS_BITS_GET(bits, BLOCK_READY);
		uint8_t block_waiting  = MARS_BITS_GET(bits, BLOCK_WAITING);
		uint8_t block_priority = MARS_BITS_GET(bits, BLOCK_PRIORITY);
		uint16_t block_counter = MARS_BITS_GET(bits, BLOCK_COUNTER);

		/* block is ready so check scheduling conditions */
		if (block_ready) {
			if (block < 0 || block_priority > max_block_priority ||
			    (block_priority == max_block_priority &&
			     block_counter > max_block_counter)) {
				block = i;
				max_block_priority = block_priority;
				max_block_counter = block_counter;
			}

			/* increment block counter */
			update_header_bits_counter(i, 0);
		}

		/* block is waiting so check block */
		if (block_waiting)
			search_block(i, 0);
	}

	/* no runnable workload found */
	if (block < 0)
		return MARS_KERNEL_STATUS_IDLE;

	/* search block for workload index to run */
	index = search_block(block, 1);
	if (index < 0)
		return MARS_KERNEL_STATUS_IDLE;

	/* set global workload info based on workload block and index */
	workload_id = MARS_WORKLOAD_PER_BLOCK * block + index;
	workload_ea = get_workload_ea(workload_id);

	/* get the workload context code from workload queue */
	dma_get(&workload, workload_ea, sizeof(struct mars_workload_context),
		MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);

	return MARS_KERNEL_STATUS_BUSY;
}

static void release_workload(void)
{
	int block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_id % MARS_WORKLOAD_PER_BLOCK;
	uint64_t block_ea = get_block_ea(block);

	/* put the workload context into workload queue */
	dma_put(&workload, workload_ea, sizeof(struct mars_workload_context),
		MARS_KERNEL_DMA_TAG);
	dma_wait(MARS_KERNEL_DMA_TAG);

	/* lock the queue block */
	mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* update current workload state in workload queue block */
	MARS_BITS_SET(&queue_block.bits[index], WORKLOAD_STATE, workload_state);

	/* update queue header bits */
	update_header_bits(block);

	/* unlock the queue block */
	mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	/* workload state is finished so notify host */
	if (workload_state == MARS_WORKLOAD_STATE_FINISHED)
		notify_host_bits(block_ea, index);
}

static void __attribute__((noinline)) run_workload(void)
{
	register void *sp asm("$sp");

	/* save kernel stack pointer */
	__kernel_stack = sp;

	/* call module entry function */
	((module_entry)workload.module.entry)(&kernel_syscalls);
}

static void load_workload_module(void)
{
	__vector unsigned char *bss_ptr, *bss_end;

	/* 0 the bss section */
	bss_ptr = (__vector unsigned char *)(MARS_WORKLOAD_MODULE_BASE_ADDR +
					     workload.module.exec_size);
	bss_end = (__vector unsigned char *)((void *)bss_ptr +
					     workload.module.bss_size);

	while (bss_ptr < bss_end)
		*bss_ptr++ = spu_splats((unsigned char)0);

	/* only reload the workload module if different from cached */
	if (kernel_memcmp(&cached_workload_module, &workload.module,
			  sizeof(struct mars_workload_module))) {
		/* load the exec code into mpu storage from host storage */
		dma_get((void *)MARS_WORKLOAD_MODULE_BASE_ADDR,
			workload.module.exec_ea, workload.module.exec_size,
			MARS_KERNEL_DMA_TAG);
		dma_wait(MARS_KERNEL_DMA_TAG);

		/* store the current cached workload module ea */
		kernel_memcpy(&cached_workload_module, &workload.module,
			      sizeof(struct mars_workload_module));
	}

	/* sync before executing loaded code */
	spu_sync();
}

static int scheduler(void)
{
	int status = reserve_workload();

	/* workload reserved */
	if (status == MARS_KERNEL_STATUS_BUSY) {
		/* load the workload module */
		load_workload_module();

		/* run workload */
		run_workload();

		/* release reservation of current workload */
		release_workload();
	}

	return status;
}

static void scheduler_idle_wait(void)
{
	int mask;
	struct mars_workload_queue_header cur_queue_header;

	/* save event mask */
	mask = spu_read_event_mask();

	/* set event mask for the lost event */
	spu_write_event_mask(MFC_LLR_LOST_EVENT);

	/* get current atomic state of queue header */
	mfc_getllar(&cur_queue_header, kernel_params.workload_queue_ea, 0, 0);
	mfc_read_atomic_status();

	/* check if queue header has been modified since we last fetched it */
	if (!kernel_memcmp(&queue_header, &cur_queue_header,
			   sizeof(struct mars_workload_queue_header))) {
		/* wait until queue header is modified */
		spu_read_event_status();
		spu_write_event_ack(MFC_LLR_LOST_EVENT);
	}

	/* restore event mask */
	spu_write_event_mask(mask);

	/* clear any remnant lost event */
	spu_write_event_ack(MFC_LLR_LOST_EVENT);
}

static void get_params(uint64_t kernel_params_ea)
{
	int mask;

	/* save event mask */
	mask = spu_read_event_mask();

	/* set event mask for the lost event */
	spu_write_event_mask(MFC_LLR_LOST_EVENT);

	/* set sync begin flag so host knows to begin sync process */
	do {
		mfc_getllar(&kernel_params, kernel_params_ea, 0, 0);
		mfc_read_atomic_status();

		/* set the flag */
		kernel_params.kernel_ticks.flag =
			MARS_KERNEL_TICKS_FLAG_SYNC_BEGIN;
		spu_dsync();

		mfc_putllc(&kernel_params, kernel_params_ea, 0, 0);
	} while (mfc_read_atomic_status() & MFC_PUTLLC_STATUS);

	/* get the tick offset from host and check if sync end flag is set */
	do {
		spu_write_event_ack(MFC_LLR_LOST_EVENT);

		mfc_getllar(&kernel_params, kernel_params_ea, 0, 0);
		mfc_read_atomic_status();

		/* host set the sync end flag so finish */
		if (kernel_params.kernel_ticks.flag ==
		    MARS_KERNEL_TICKS_FLAG_SYNC_END)
			break;

		spu_read_event_status();
	} while (1);

	/* now, kernel parameters, including offset of tick counters,
	 * are stored in 'kernel_params'.
	 */

	/* start decrementer */
	spu_write_decrementer(0);

	/* restore event mask */
	spu_write_event_mask(mask);

	/* clear any remnant lost event */
	spu_write_event_ack(MFC_LLR_LOST_EVENT);
}

int main(unsigned long long mpu_context_id,
	 unsigned long long kernel_params_ea)
{
	(void)mpu_context_id;

	int exit_flag = 0;

	get_params(kernel_params_ea);

	while (!exit_flag) {
		int status = scheduler();

		switch (status) {
		case MARS_KERNEL_STATUS_BUSY:
			break;
		case MARS_KERNEL_STATUS_IDLE:
			scheduler_idle_wait();
			break;
		case MARS_KERNEL_STATUS_EXIT:
			exit_flag = 1;
			break;
		}
	}

	host_signal_send(MARS_HOST_SIGNAL_EXIT);

	return MARS_SUCCESS;
}
