/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <spu_mfcio.h>

#include <mars/error.h>
#include <mars/module.h>

#include "config.h"

#include "task_module.h"

#define MARS_TASK_MODULE_DMA_SIZE_MAX		16384
#define MARS_TASK_MODULE_DMA_SIZE_MASK		0x7f

#define MARS_TASK_REGISTER_SAVE_AREA_SIZE 	(16 * (127 - 80 + 1))

/* stack pointer storage */
void *__module_stack;
void *__task_stack;

/* global task variables */
static struct mars_task_context *task;
static int task_cached;

/* task entry */
typedef void (*mars_task_entry)(struct mars_task_args *args,
	struct mars_task_module_syscalls *task_module_syscalls);

static uint32_t get_ticks(void)
{
	return mars_module_get_ticks();
}

static uint16_t get_kernel_id(void)
{
	return mars_module_get_kernel_id();
}

static struct mars_task_context *get_task(void)
{
	return (struct mars_task_context *)mars_module_get_workload();
}

static struct mars_task_context *get_task_by_id(struct mars_task_id *task_id)
{
	return (struct mars_task_context *)
		mars_module_get_workload_by_id(task_id->workload_id);
}

static void registers_save(void)
{
	/* push non-volatile registers on to the workload stack */
	asm volatile (
		"stqd	$80, -16(%[ptr]);"
		"stqd	$81, -32(%[ptr]);"
		"stqd	$82, -48(%[ptr]);"
		"stqd	$83, -64(%[ptr]);"
		"stqd	$84, -80(%[ptr]);"
		"stqd	$85, -96(%[ptr]);"
		"stqd	$86, -112(%[ptr]);"
		"stqd	$87, -128(%[ptr]);"
		"stqd	$88, -144(%[ptr]);"
		"stqd	$89, -160(%[ptr]);"
		"stqd	$90, -176(%[ptr]);"
		"stqd	$91, -192(%[ptr]);"
		"stqd	$92, -208(%[ptr]);"
		"stqd	$93, -224(%[ptr]);"
		"stqd	$94, -240(%[ptr]);"
		"stqd	$95, -256(%[ptr]);"
		"stqd	$96, -272(%[ptr]);"
		"stqd	$97, -288(%[ptr]);"
		"stqd	$98, -304(%[ptr]);"
		"stqd	$99, -320(%[ptr]);"
		"stqd	$100, -336(%[ptr]);"
		"stqd	$101, -352(%[ptr]);"
		"stqd	$102, -368(%[ptr]);"
		"stqd	$103, -384(%[ptr]);"
		"stqd	$104, -400(%[ptr]);"
		"stqd	$105, -416(%[ptr]);"
		"stqd	$106, -432(%[ptr]);"
		"stqd	$107, -448(%[ptr]);"
		"stqd	$108, -464(%[ptr]);"
		"stqd	$109, -480(%[ptr]);"
		"stqd	$110, -496(%[ptr]);"
		"stqd	$111, -512(%[ptr]);"
		"stqd	$112, -528(%[ptr]);"
		"stqd	$113, -544(%[ptr]);"
		"stqd	$114, -560(%[ptr]);"
		"stqd	$115, -576(%[ptr]);"
		"stqd	$116, -592(%[ptr]);"
		"stqd	$117, -608(%[ptr]);"
		"stqd	$118, -624(%[ptr]);"
		"stqd	$119, -640(%[ptr]);"
		"stqd	$120, -656(%[ptr]);"
		"stqd	$121, -672(%[ptr]);"
		"stqd	$122, -688(%[ptr]);"
		"stqd	$123, -704(%[ptr]);"
		"stqd	$124, -720(%[ptr]);"
		"stqd	$125, -736(%[ptr]);"
		"stqd	$126, -752(%[ptr]);"
		"stqd	$127, -768(%[ptr]);"
		: : [ptr] "r" (__task_stack));
}

static void registers_restore(void)
{
	/* pop non-volatile registers from saved workload stack */
	asm volatile (
		"lqd	$80, -16(%[ptr]);"
		"lqd	$81, -32(%[ptr]);"
		"lqd	$82, -48(%[ptr]);"
		"lqd	$83, -64(%[ptr]);"
		"lqd	$84, -80(%[ptr]);"
		"lqd	$85, -96(%[ptr]);"
		"lqd	$86, -112(%[ptr]);"
		"lqd	$87, -128(%[ptr]);"
		"lqd	$88, -144(%[ptr]);"
		"lqd	$89, -160(%[ptr]);"
		"lqd	$90, -176(%[ptr]);"
		"lqd	$91, -192(%[ptr]);"
		"lqd	$92, -208(%[ptr]);"
		"lqd	$93, -224(%[ptr]);"
		"lqd	$94, -240(%[ptr]);"
		"lqd	$95, -256(%[ptr]);"
		"lqd	$96, -272(%[ptr]);"
		"lqd	$97, -288(%[ptr]);"
		"lqd	$98, -304(%[ptr]);"
		"lqd	$99, -320(%[ptr]);"
		"lqd	$100, -336(%[ptr]);"
		"lqd	$101, -352(%[ptr]);"
		"lqd	$102, -368(%[ptr]);"
		"lqd	$103, -384(%[ptr]);"
		"lqd	$104, -400(%[ptr]);"
		"lqd	$105, -416(%[ptr]);"
		"lqd	$106, -432(%[ptr]);"
		"lqd	$107, -448(%[ptr]);"
		"lqd	$108, -464(%[ptr]);"
		"lqd	$109, -480(%[ptr]);"
		"lqd	$110, -496(%[ptr]);"
		"lqd	$111, -512(%[ptr]);"
		"lqd	$112, -528(%[ptr]);"
		"lqd	$113, -544(%[ptr]);"
		"lqd	$114, -560(%[ptr]);"
		"lqd	$115, -576(%[ptr]);"
		"lqd	$116, -592(%[ptr]);"
		"lqd	$117, -608(%[ptr]);"
		"lqd	$118, -624(%[ptr]);"
		"lqd	$119, -640(%[ptr]);"
		"lqd	$120, -656(%[ptr]);"
		"lqd	$121, -672(%[ptr]);"
		"lqd	$122, -688(%[ptr]);"
		"lqd	$123, -704(%[ptr]);"
		"lqd	$124, -720(%[ptr]);"
		"lqd	$125, -736(%[ptr]);"
		"lqd	$126, -752(%[ptr]);"
		"lqd	$127, -768(%[ptr]);"
		: : [ptr] "r" (__task_stack));
}

static void dma_context(int save, void *task_heap)
{
	uint32_t low_size, high_size;

	/* save or restore context */
	if (save) {
		/* text, data and heap (low address) */
		low_size =
			(((uintptr_t)task_heap -
			  MARS_TASK_BASE_ADDR +
			  MARS_TASK_MODULE_DMA_SIZE_MASK) &
			 ~MARS_TASK_MODULE_DMA_SIZE_MASK);

		/* stack (high address) */
		high_size =
			((MARS_TASK_BASE_ADDR +
			  MARS_TASK_CONTEXT_SAVE_SIZE_MAX -
			  (uintptr_t)__task_stack +
			  MARS_TASK_REGISTER_SAVE_AREA_SIZE +
			  MARS_TASK_MODULE_DMA_SIZE_MASK) &
			 ~MARS_TASK_MODULE_DMA_SIZE_MASK);

		/* full context save/restore */
		if (low_size + high_size >= MARS_TASK_CONTEXT_SAVE_SIZE_MAX) {
			low_size = MARS_TASK_CONTEXT_SAVE_SIZE_MAX;
			high_size = 0;
		}

		/* save save sizes to task context */
		task->context_save_area_low_size = low_size;
		task->context_save_area_high_size = high_size;

		/* save text and heap (low address) */
		mars_module_dma_put((void *)MARS_TASK_BASE_ADDR,
				task->context_save_area_ea,
				low_size, MARS_TASK_MODULE_DMA_TAG);

		/* save stack (high address) */
		mars_module_dma_put((void *)(MARS_TASK_BASE_ADDR +
				MARS_TASK_CONTEXT_SAVE_SIZE_MAX - high_size),
				task->context_save_area_ea + low_size,
				high_size, MARS_TASK_MODULE_DMA_TAG);
	} else {
		/* restore save sizes from task context */
		low_size = task->context_save_area_low_size;
		high_size = task->context_save_area_high_size;

		/* restore text and heap (low address) */
		mars_module_dma_get((void *)MARS_TASK_BASE_ADDR,
				task->context_save_area_ea,
				low_size, MARS_TASK_MODULE_DMA_TAG);

		/* restore stack (high address) */
		mars_module_dma_get((void *)(MARS_TASK_BASE_ADDR +
				MARS_TASK_CONTEXT_SAVE_SIZE_MAX - high_size),
				task->context_save_area_ea + low_size,
				high_size, MARS_TASK_MODULE_DMA_TAG);
	}

	mars_module_dma_wait(MARS_TASK_MODULE_DMA_TAG);
}

static void __attribute__((noinline)) context_save(void *task_heap)
{
	/* save registers state */
	registers_save();

	/* save workload stack pointer */
	task->stack = (uint32_t)__task_stack;

	/* save context MPU storage state */
	dma_context(1, task_heap);
}

static void __attribute__((noinline)) context_restore(void)
{
	/* if task not cashed restore context MPU storage state */
	if (!task_cached)
		dma_context(0, NULL);

	/* restore workload stack pointer */
	__task_stack = (void *)task->stack;

	/* restore registers state */
	registers_restore();
}

static void __attribute__((noinline)) context_wait(void *task_heap)
{
	register void *sp asm("$sp");

	/* save task stack pointer */
	__task_stack = sp;

	/* restore module stack pointer */
	sp = __module_stack;

	/* save task context */
	context_save(task_heap);

	/* put workload into waiting state */
	mars_module_workload_wait();
}

static void __attribute__((noinline)) context_resume(void)
{
	register void *sp asm("$sp");

	/* restore task context */
	context_restore();

	/* restore task stack pointer */
	sp = __task_stack;

	/* sync before executing loaded code */
	spu_sync();
}

static void task_yield(void *task_heap)
{
	register void *sp asm("$sp");

	/* save task stack pointer and restore module stack pointer */
	__task_stack = sp;
	sp = __module_stack;

	/* save task context */
	context_save(task_heap);

	/* yield workload and put into ready state */
	mars_module_workload_yield();
}

static void task_exit(void)
{
	/* restore module stack pointer */
	register void *sp asm("$sp");

	sp = __module_stack;

	/* put workload into finished state */
	mars_module_workload_finish();
}

static int task_schedule(uint16_t workload_id, struct mars_task_args *args,
			 uint8_t priority)
{
	int ret;
	struct mars_task_context *schedule_task;
	struct mars_workload_context *schedule_workload;

	ret = mars_module_workload_schedule_begin(workload_id, priority,
						  &schedule_workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	schedule_task = (struct mars_task_context *)schedule_workload;

	/* initialize task specific context variables */
	schedule_task->stack = 0;
	if (args)
		schedule_task->args = *args;

	/* end process to schedule the workload in the workload queue */
	return mars_module_workload_schedule_end(workload_id);
}

static int task_wait(uint16_t workload_id, void *task_heap)
{
	int ret;

	ret = mars_module_workload_wait_set(workload_id);
	if (ret != MARS_SUCCESS)
		return ret;

	context_wait(task_heap);

	return mars_module_workload_wait_reset();
}

static int task_try_wait(uint16_t workload_id)
{
	/* make sure workload is initialized */
	if (!mars_module_workload_query(workload_id, MARS_QUERY_IS_INITIALIZED))
		return MARS_ERROR_STATE;

	/* if workload not finished return busy */
	if (!mars_module_workload_query(workload_id, MARS_QUERY_IS_FINISHED))
		return MARS_ERROR_BUSY;

	return mars_module_workload_wait_reset();
}

static int task_signal_host(uint64_t watch_point_ea)
{
	return mars_module_host_signal_send(watch_point_ea);
}

static int task_signal_send(uint16_t workload_id)
{
	return mars_module_workload_signal_set(workload_id);
}

static int task_signal_wait(void *task_heap)
{
	context_wait(task_heap);

	return mars_module_workload_signal_reset();
}

static int task_signal_try_wait(void)
{
	/* if signal not yet received return busy */
	if (!mars_module_workload_query(mars_module_get_workload_id(),
					MARS_QUERY_IS_SIGNAL_SET))
		return MARS_ERROR_BUSY;

	return mars_module_workload_signal_reset();
}

static struct mars_task_module_syscalls task_module_syscalls =
{
	get_ticks,
	get_kernel_id,
	get_task,
	get_task_by_id,

	task_exit,
	task_yield,
	task_schedule,
	task_wait,
	task_try_wait,
	task_signal_host,
	task_signal_send,
	task_signal_wait,
	task_signal_try_wait,

	mars_module_mutex_lock_get,
	mars_module_mutex_unlock_put,

	mars_module_dma_get,
	mars_module_dma_put,
	mars_module_dma_wait
};

static void context_start(void)
{
	__vector unsigned char *bss_ptr, *bss_end;

	/* 0 the bss section */
	bss_ptr = (__vector unsigned char *)
			(MARS_TASK_BASE_ADDR + task->exec_size);
	bss_end = (__vector unsigned char *)
			((void *)bss_ptr + task->bss_size);

	while (bss_ptr < bss_end)
		*bss_ptr++ = spu_splats((unsigned char)0);

	/* dma the text and data section */
	mars_module_dma_get((void *)MARS_TASK_BASE_ADDR, task->exec_ea,
			    task->exec_size, MARS_TASK_MODULE_DMA_TAG);
	mars_module_dma_wait(MARS_TASK_MODULE_DMA_TAG);

	/* sync before executing loaded code */
	spu_sync();

	/* call entry function */
	((mars_task_entry)task->entry)(&task->args, &task_module_syscalls);
}

void mars_module_main(void)
{
	register void *sp asm("$sp");

	/* save module stack pointer */
	__module_stack = sp;

	/* get task context */
	task = get_task();

	/* check if task context is cached in mpu storage */
	task_cached = mars_module_workload_query(mars_module_get_workload_id(),
						 MARS_QUERY_IS_CACHED);

	/* if stack pointer is uninitialized run fresh, otherwise resume */
	if (!task->stack)
		context_start();
	else
		context_resume();

	/* we should never reach here */
}
